#ifndef __MEMETYPE__
#define __MEMETYPE__

#include "memebase.h"

#include <string>
#include <vector>
#include <map>

using namespace std;

class MemeString;
class Value;

class MemePrimitive : public MemeCollect {
	public:
		// these are not abstract, they all throw
		// errors if called on an object that does
		// not implement
		MemePrimitive();
		~MemePrimitive();

		virtual Value ini(const Value& x);

		virtual Value add(const Value& a, const Value& b);
		virtual Value sub(const Value& a, const Value& b);
		virtual Value neg(const Value& x);

		virtual Value mul(const Value& a, const Value& b);
		virtual Value div(const Value& a, const Value& b);
		virtual Value mod(const Value& a, const Value& b);

		virtual Value cmp(const Value& a, const Value& b);

		virtual Value len(const Value& a);
		virtual Value idx(const Value& a, const Value& i);
		virtual Value ins(const Value& a, const Value& b);
		virtual Value cat(const Value& a, const Value& b);

		virtual Value str(const Value& a);
		virtual Value num(const Value& a);

		// this is what is called on lvalues to assign to them,
		// do not implement for constant values.
		// lvalues are always references, yes it's const.
		virtual void lassign(const Value& left, const Value& right);
};


typedef enum {
	valueNull=0,
	valueInt,
	valueReal,
	valueBool,
	valueStr,
	valueArray,
	valueHash,
	valueRef,
	valueAppender,
	valueFunc
} valueEnum;

// these are not always gc'ed but we have gc setup so you can manage things
// yourself as long as you never create a ref to it.
class Value : public MemeCollect {
	public:
	valueEnum type;

	union {
		int i;
		bool b;
		double r;
	};
	MemeRef ref_; // won't let me put this in union so we're wasting space

	Value(void);
	Value(int v);
	Value(double v);
	Value(bool v);
	Value(valueEnum t, const MemeRef& ref);

	Value(const Value& v );


	const Value& operator = (const Value& x );

	void assign(const Value& val);

	int toInt(void) const;
	double toReal(void) const;
	const char *toString(void) const;
	bool toBool(void) const;

	Value value(void) const;

	void print(void) const;
	void println(void) const;

	static void registerPrimitive(valueEnum type, MemePrimitive* prim);
	MemePrimitive* getPrimitive(void) const;
} ;

inline Value operator + (const Value& a, const Value& b )
{
	return a.getPrimitive()->add( a, b );
}

inline Value operator - (const Value& a, const Value& b ) {
	return a.getPrimitive()->sub( a, b );
}

inline Value operator - (const Value& a) {
	return a.getPrimitive()->neg( a );
}

inline Value operator * (const Value& a, const Value& b )
{
	return a.getPrimitive()->mul( a, b );
}

inline Value operator / (const Value& a, const Value& b )
{
	return a.getPrimitive()->div( a, b );
}

inline Value operator % (const Value& a, const Value& b )
{
	return a.getPrimitive()->mod( a, b );
}

inline bool operator < (const Value& a, const Value& b ) {
	return a.getPrimitive()->cmp( a, b ).i < 0;
}

inline bool operator > (const Value& a, const Value& b ) {
	return a.getPrimitive()->cmp( a, b ).i > 0;
}
inline bool operator <= (const Value& a, const Value& b ) {
	return a.getPrimitive()->cmp( a, b ).i <= 0;
}
inline bool operator >= (const Value& a, const Value& b ) {
	return a.getPrimitive()->cmp( a, b ).i >= 0;
}

inline bool operator == (const Value& a, const Value& b ) {
	return a.getPrimitive()->cmp( a, b ).i == 0;
}

inline bool operator != (const Value& a, const Value& b ) {
	return a.getPrimitive()->cmp( a, b ).i != 0;
}



typedef std::pair<std::string,Value> SymEntry;

// one of these for each set of local variables
// and for each non-intrsic type with data members
class MemeScope : public MemeCollect {
	private:
		std::map<std::string,Value> symbols_;
		MemeScope *parent_;
		Value return_;
		bool hasReturned_, hasBreak_;
	public:
		MemeScope(MemeScope *parent);
		~MemeScope();

		void dump(void); // dump to stdout for debugging

		// define a new variable, will replace anything at that name in the scope
		MemeRef define(const char *name, const Value& val);
		// bind a ref to a variable
		void bind(const char *name, MemeRef theRef);
		// assign to variable
		Value assign(const char *name, const Value& val);
		// lookup variable
		Value* lookup(const char *name);

		bool unwind( void ) const;
		// we use scope to implement stack frames, and return and
		// other types of unwinding.
		// this call is used to signal a return.
		void func_return(const Value& val);

		void set_break(void);
		void clr_break(void);

		Value getReturn(void) const;
};

class MemeString : public MemeCollect {
	private:
		std::string str_;
	public:
		MemeString(void);
		MemeString(const char *str);
		MemeString(const char *str, int len);
		MemeString(const MemeString& ms);
		~MemeString();

		const MemeString& operator = (const MemeString& ms);

		const char *getText(void) const;
		int getLength(void) const;
		void setText( const char *str );

		void append( const char *str );

		// create and store a string and return a reference to it.
		static MemeRef newString( const char *str );
		static MemeRef newString( const char *str, int len );

		// get string from reference
		static MemeString* derefString( const MemeRef& ref );
		static const char* strFromRef( const MemeRef& ref );
};

struct nodeTypeTag;
class MemeFunc : public MemeCollect {
	protected:
		struct nodeTypeTag *p_;
		MemeScope *parent_;

		void SetupLocals(std::vector<Value> & args, MemeScope & locals);
	public:
		MemeFunc(MemeScope *parent, struct nodeTypeTag *p);
		~MemeFunc();

		virtual Value call( std::vector<Value>& args ) = 0;
};

typedef Value (*MemeCppProc)(MemeScope& scope);


class MemeNativeFunc : public MemeFunc {
public:
	MemeNativeFunc(MemeScope *parent, struct nodeTypeTag *p);
	~MemeNativeFunc();

	virtual Value call( std::vector<Value>& args );
};

class MemeCppFunc : public MemeFunc {
private:
	MemeCppProc proc_;
public:
	MemeCppFunc(MemeScope *parent, struct nodeTypeTag *p, MemeCppProc proc);
	~MemeCppFunc();

	virtual Value call( std::vector<Value>& args );
};


void RegisterCppFunc(MemeScope& scope, const char *name, std::string paramNames, MemeCppProc proc);
void RegisterConstant(MemeScope& scope, const char *name, const Value& v);

/* one day ...
class MemeType : public MemePrimitive {
	private:
		MemeType *super_;
		MemeScope mesg_;
	public:
};

class MemeInstance : public MemeCollect {
	private:
		MemeType *type_;
		MemeScope self_;
	public:

};
*/

#endif


